@@ -332,7 +332,7 @@ class Individual:
         if reset_parent_name_list:
             self.parents = list()
         for parent in parents:
-            self.parents.append(parent.name)
+            self.parents.append(parent)
 
     def reset_name(self, name : str = None):
         """
@@ -558,8 +558,20 @@ class Island:
         children (list): All children created during the run.
         generation_count (int): The current generation number.
         checkpoints_dir (str): Directory name of where all checkpoints are saved.
+        lineage (dict): A graph of lineage for the whole population.
     """
 
+    __stat_key = 'stat'
+    __generation_key = 'generation'
+    __pop_len_key = 'pop_len'
+    __fitness_mean_key = 'fitness_mean'
+    __fitness_std_key = 'fitness_std'
+    __fitness_min_key = 'fitness_min'
+    __fitness_max_key = 'fitness_max'
+    __most_fit_key = 'most_fit'
+    __least_fit_key = 'least_fit'
+
+
     def __init__(self, function_params : dict = None,
                  initialisation_function: Callable = initialise_population_random,
                  elite_selection : Callable = selection_elites_top_n,
@@ -602,6 +614,10 @@ class Island:
             self.function_params = dict()
         self.unique_genome = list()
         self.generation_info = list()
+        self.lineage = {
+            'nodes' : [],
+            'edges' : []
+        }
         self.population = list()
 
         self._initialise = initialisation_function
@@ -776,7 +792,7 @@ class Island:
             _initialisation_params = {}
 
         self.verbose_logging(f"init: pop_size {population_size}")
-        self.verbose_logging(f"init: adam {repr(adam)}")
+        self.verbose_logging(f"init: adam {str(adam)}")
         self.population = self._initialise(adam=adam, n=population_size, island=self, **_initialisation_params)
 
         if self.save_checkpoint_level == 2:
@@ -784,13 +800,25 @@ class Island:
 
         if evaluate_population:
             for popitem in self.population:
-                self.verbose_logging(f"init: eval {repr(popitem)}")
+                self.verbose_logging(f"init: eval {str(popitem)}")
                 popitem.evaluate(self.function_params, island=self)
                 self.unique_genome.append(popitem.unique_genetic_code())
         if self.save_checkpoint_level == 2:
             self.save_checkpoint(event='init_post', island=self)
         self.verbose_logging("init: complete")
 
+        for p in self.population:
+            self.__add_to_lineage(p)
+
+    def __add_to_lineage(self, individual : Individual):
+        self.lineage['nodes'].append({
+            'name': individual.name,
+            'age': individual.age,
+            'fitness': individual.fitness,
+            'chromosome': str(individual.chromosome),
+            '_individual' : individual
+        })
+
     def import_migrants(self, migrants : list,
                         reset_fitness : bool = False,
                         species_check : bool = True,
@@ -806,25 +834,28 @@ class Island:
             force_genetic_diversity (bool): Only add migrants to the population if they have a unique chromosome (default = True).
         """
         for i in migrants:
-            self.verbose_logging(f"migration: customs {repr(i)}")
+            self.verbose_logging(f"migration: customs {str(i)}")
             if species_check:
                 if i.species_type != self.species_type:
                     continue
             if force_genetic_diversity:
                 if not i.unique_genetic_code() in self.unique_genome:
                     if i.fitness is None or reset_fitness:
-                        self.verbose_logging(f"migration: eval {repr(i)}")
+                        self.verbose_logging(f"migration: eval {str(i)}")
                         i.evaluate(self.function_params, island=self)
-                    self.verbose_logging(f"migration: add {repr(i)}")
+                    self.verbose_logging(f"migration: add {str(i)}")
                     self.population.append(i)
                     self.unique_genome.append(i.unique_genetic_code())
+                    self.__add_to_lineage(i)
             else:
                 if i.fitness is None or reset_fitness:
-                    self.verbose_logging(f"migration: eval {repr(i)}")
+                    self.verbose_logging(f"migration: eval {str(i)}")
                     i.evaluate(self.function_params)
-                self.verbose_logging(f"migration: add {repr(i)}")
+                self.verbose_logging(f"migration: add {str(i)}")
                 self.population.append(i)
                 self.unique_genome.append(i.unique_genetic_code())
+                self.__add_to_lineage(i)
+
         self.verbose_logging("migration: imported")
 
 
@@ -922,7 +953,7 @@ class Island:
         best_ind = selection_elites_top_n(island=self, individuals=self.population, n=1)[0]
 
         self.verbose_logging(f"evolve: end")
-        self.verbose_logging(f"evolve: best {repr(best_ind)}")
+        self.verbose_logging(f"evolve: best {str(best_ind)}")
 
         if post_evolution_function:
             post_evolution_function(island=self)
@@ -976,14 +1007,14 @@ class Island:
         for parents in self.parent_selection(individuals=elites, island=self, **parent_selection_params):
             self.verbose_logging(f"select: parent_count {len(parents)}")
             if self.crossover_prob(crossover_probability=crossover_probability, island=self):
-                self.verbose_logging(f"cross: parents {[repr(p) for p in parents]}")
+                self.verbose_logging(f"cross: parents {[str(p) for p in parents]}")
 
                 children = self.crossover(island=self,
                                           individuals=self.clone(individuals=parents, island=self),
                                           **crossover_params)
 
                 self.verbose_logging(f"cross: offspring_count {len(children)}")
-                self.verbose_logging(f"cross: offspring {[repr(p) for p in children]}")
+                self.verbose_logging(f"cross: offspring {[str(p) for p in children]}")
 
                 for child in children:
                     child.reset_name()
@@ -998,7 +1029,7 @@ class Island:
         generation_mutants = list()
         for mutant in generation_children:
             if self.mutation_prob(mutation_probability=mutation_probability, island=self):
-                self.verbose_logging(f"mutate: offspring {repr(mutant)}")
+                self.verbose_logging(f"mutate: offspring {str(mutant)}")
 
                 mutated = self.mutation(island=self, individual=mutant, **mutation_params)
 
@@ -1012,22 +1043,30 @@ class Island:
             self.save_checkpoint(event=f'evolve_pre_eval_{g}', island=self)
         for individual in generation_children:
             individual.reset_fitness()
-            self.verbose_logging(f"evolve: eval {repr(individual)}")
+            self.verbose_logging(f"evolve: eval {str(individual)}")
             individual.evaluate(island=self, params=self.function_params)
             offspring_fitnesses.append(individual.fitness)
         if self.save_checkpoint_level == 2:
             self.save_checkpoint(event=f'evolve_post_eval_{g}', island=self)
 
+        for child in generation_children:
+            self.__add_to_lineage(child)
+            for p in child.parents:
+                self.lineage['edges'].append({
+                    'from': p.name,
+                    'to': child.name,
+                    'generation' : g
+                })
 
         for individual in self.survivor_selection(individuals=generation_children, island=self, **survivor_selection_params):
             if self.allow_twins:
                 # Else, add it effectively allowing "twins" to exist
-                self.verbose_logging(f"evolve: add {repr(individual)}")
+                self.verbose_logging(f"evolve: add {str(individual)}")
                 self.population.append(individual)
                 self.unique_genome.append(individual.unique_genetic_code())
             elif not individual.unique_genetic_code() in self.unique_genome:
                 # If we want a diverse gene pool, this must be true
-                self.verbose_logging(f"evolve: add {repr(individual)}")
+                self.verbose_logging(f"evolve: add {str(individual)}")
                 self.population.append(individual)
                 self.unique_genome.append(individual.unique_genetic_code())
 
@@ -1038,26 +1077,30 @@ class Island:
         if len(offspring_fitnesses) > 0:
             self.generation_info.append(
                 {
-                    'stat': 'offspring',
-                    'generation': g,
-                    'pop_len': len(offspring_fitnesses),
-                    'fitness_mean': np.mean(offspring_fitnesses),
-                    'fitness_std': np.std(offspring_fitnesses),
-                    'fitness_min': min(offspring_fitnesses),
-                    'fitness_max': max(offspring_fitnesses),
+                    self.__generation_key: g,
+                    self.__stat_key: 'offspring',
+                    self.__pop_len_key: len(offspring_fitnesses),
+                    self.__fitness_mean_key: np.mean(offspring_fitnesses),
+                    self.__fitness_std_key: np.std(offspring_fitnesses),
+                    self.__fitness_min_key: min(offspring_fitnesses),
+                    self.__fitness_max_key: max(offspring_fitnesses),
+                    self.__most_fit_key: selection_elites_top_n(island=self, individuals=generation_children, n=1)[0].name,
+                    self.__least_fit_key : selection_elites_top_n(island=self, individuals=generation_children, n=1)[-1].name
                 }
             )
             self.verbose_logging(f"evolve: stats {self.generation_info[-1]}")
 
         self.generation_info.append(
             {
-                'stat': 'elites',
-                'generation': g,
-                'pop_len': len(elite_fitnesses),
-                'fitness_mean': np.mean(elite_fitnesses),
-                'fitness_std': np.std(elite_fitnesses),
-                'fitness_min': min(elite_fitnesses),
-                'fitness_max': max(elite_fitnesses),
+                self.__generation_key: g,
+                self.__stat_key: 'elites',
+                self.__pop_len_key: len(elite_fitnesses),
+                self.__fitness_mean_key: np.mean(elite_fitnesses),
+                self.__fitness_std_key: np.std(elite_fitnesses),
+                self.__fitness_min_key: min(elite_fitnesses),
+                self.__fitness_max_key: max(elite_fitnesses),
+                self.__most_fit_key: selection_elites_top_n(island=self, individuals=elites, n=1)[0].name,
+                self.__least_fit_key : selection_elites_top_n(island=self, individuals=elites, n=1)[-1].name
             }
         )
 
@@ -1065,13 +1108,15 @@ class Island:
 
         self.generation_info.append(
             {
-                'stat': 'population',
-                'generation': g,
-                'pop_len': len(population_fitnesses),
-                'fitness_mean': np.mean(population_fitnesses),
-                'fitness_std': np.std(population_fitnesses),
-                'fitness_min': min(population_fitnesses),
-                'fitness_max': max(population_fitnesses),
+                self.__generation_key: g,
+                self.__stat_key: 'population',
+                self.__pop_len_key: len(population_fitnesses),
+                self.__fitness_mean_key: np.mean(population_fitnesses),
+                self.__fitness_std_key: np.std(population_fitnesses),
+                self.__fitness_min_key: min(population_fitnesses),
+                self.__fitness_max_key: max(population_fitnesses),
+                self.__most_fit_key: selection_elites_top_n(island=self, individuals=self.population, n=1)[0].name,
+                self.__least_fit_key: selection_elites_top_n(island=self, individuals=self.population, n=1)[-1].name
             }
         )
 
@@ -1085,6 +1130,58 @@ class Island:
         if self.save_checkpoint_level == 1:
             self.save_checkpoint(event=f'evolve_post_{g}', island=self)
 
+    def write_lineage_json(self, filename : str):
+        """
+        Dumps the lineage safely to JSON file.
+
+        Args:
+            filename (str): Output file.
+        """
+        nodes = list()
+        for n in self.lineage['nodes']:
+            nodes.append({
+                'name': n['name'],
+                'age': n['age'],
+                'fitness': n['fitness'],
+                'chromosome': n['chromosome'],
+            })
+        import json
+        with open(filename, 'w', newline='', encoding='utf8') as output_file:
+            json.dump({'nodes' : nodes, 'edges' : self.lineage['edges']}, output_file)
+
+
+    def write_report(self, filename : str, output_json : bool = False):
+        """
+        Write the generational history to CSV (or JSON) file.
+
+        Args:
+            filename (str): Output file.
+            output_json (bool): Write as JSON instead of CSV (default = False).
+        """
+        if output_json:
+            import json
+            with open(filename, 'w', newline='', encoding='utf8') as output_file:
+                json.dump(self.generation_info, output_file)
+        else:
+            import csv
+            keys = [
+                self.__generation_key,
+                self.__stat_key,
+                self.__pop_len_key,
+                self.__fitness_mean_key,
+                self.__fitness_std_key,
+                self.__fitness_min_key,
+                self.__fitness_max_key,
+                self.__most_fit_key,
+                self.__least_fit_key
+            ]
+            with open(filename, 'w', newline='', encoding='utf8') as output_file:
+                dict_writer = csv.DictWriter(output_file, keys)
+                dict_writer.writeheader()
+                dict_writer.writerows(self.generation_info)
+
+
+
     def verbose_logging(self, event_message):
         if self.verbose:
             logging.info(event_message, extra={'island' : self.name})
